function [logL,Ps] = thresholdPosterior(T,C,theta,N)
% [logL,Ps] = thresholdPosterior(T,C,theta)
%
% Compute (approximate) log Likelihood of the characters in C on the tree
% T under Wright & Felsensteins threshold model as well as posterior
% densities at each node. This algorithm uses a Gaussian kernel method.
%
% T contains the following fields:
%   T.children  nnodes x 2 array specifying children of each node (or 0 if
%       none)
%   T.parent nnodes x 1 array specifying parents of each node (or 0 if
%       root)
%   T.t  nnodes x 1 array specifying branch lengths. (0 if root)
% The procedure readTree can set up this structure from a NEWICK formatted
% string.
%
% C is an nleaves x nchars array. This is coding 0-1 characters, with
%   negatives denoting missing data.
%
% theta is a bundle of parameters with the following fields
%   theta.mu_r   mean of root density
%   theta.sigma2_r   variance of root density
%   theta.sigma2    variance of transition densities (per unit branch
%       length)
% If theta.sigma2_r = 0, we assume that the liability at the root is fixed
% at theta.mu_r
%
% N is the number of segments per interval (N >0)
%
% Ps is a 3 x 1 structure. Ps(1).array is a nchars x (nleaves-1) x (N+1)
% matrix P, with P(c,i,:) denoting the density for node i, character c.
% Ps(2).array is a nchars x (nleaves-1) matrix P2, with P2(c,i) denoting
% the probability that the liability of character c at node i is positive.
% Ps(3).array is a nchars x (N+1) matrix X, where X(i,:) is the set of
% liability values at node i used in the numerical integrations.

nnodes = size(T.children,1);
nchar = size(C,2);

n = (nnodes+1)/2;
x = (0:N)/N; %array used to calculate the points where the functions are evaluated
p = zeros(nnodes,1); %p(i) = length of path from root to node i
for i = (nnodes-1):-1:1
    p(i) = p(T.parent(i)) + T.t(i);
end
L = zeros(nnodes,1); %lower integration limits for each node's integral
U = zeros(nnodes,1); %upper integration limits for each node's integral
X = zeros(nnodes,N+1);
P = zeros(nchar,n-1,N+1);
P2 = zeros(nchar,n-1);
P2(:,1:n) = C'; %if leaf i has a value of 1 for character c, then
                %P2(c,i) = 1. If value is 0, then P2(c,i) = 0.

alpha = n*N^(-4)/(2*(n-1)); %truncation error of each individual integral.
for i=(n+1):nnodes
    L(i) = InvPhi(alpha,theta.mu_r,theta.sigma2_r+theta.sigma2*p(i));
    U(i) = InvPhi(1-alpha,theta.mu_r,theta.sigma2_r+theta.sigma2*p(i));
    X(i,:) = (U(i) - L(i))*x+L(i);
end

logL = 0.0;  %log of the product of the likelihoods
like = zeros(nchar,1); %like(c) = likelihood of the characters in C(c,:)

for ch = 1:nchar
    z = C(:,ch); %trait values observed at the leaves for character ch
    F = ones(nnodes,N+1); %F(i,x) = likelihood of the data at or directly
                          %below node i, given a liability of x at i's
                          %parent node.
    for i=1:n
        j = T.parent(i); %parent node of i.
        for k=0:N
            if z(i) == 1
                %integrates over positive liabilities
                F(i,k+1) = 1 - Phi(0,X(j,k+1),theta.sigma2*T.t(i));
            elseif z(i) == 0
                %integrates over negative liabilities
                F(i,k+1) = Phi(0,X(j,k+1),theta.sigma2*T.t(i));
            end
        end
    end
    for i=(n+1):(2*n-2)
        j = T.parent(i);
        u = T.children(i,1); v = T.children(i,2); %child nodes of node i.
        if T.t(i) == 0
            F(i,:) = F(u,:).*F(v,:);
        else
            %Integral calculated via Gaussian kernel method
            weight = findWeightsSet(X(i,:),X(j,:),theta.sigma2*T.t(i));
            for k=0:N
                f = F(u,:).*F(v,:);
                F(i,k+1) = sum(weight(k+1,:).*f);
            end
        end
    end
    r = 2*n-1;
    u = T.children(r,1); v = T.children(r,2);
    if theta.sigma2_r > 0
        %Integral calculated via Gaussian kernel method
        weight = findWeights(X(r,:),theta.mu_r,theta.sigma2_r);
        f = F(u,:).*F(v,:);
        like(ch)= sum(weight.*f);
    else
        %We have L_r = U_r, so X_r[k] is constant in k. 
        like(ch) = F(u,1)*F(v,1);
    end
    
    
    G = ones(nnodes,N+1);
    for i = (2*n-2):-1:(n+1)
        j = T.parent(i);
        if T.children(j,1) == i    %v = sibling of i
            v = T.children(j,2);
        else
            v = T.children(j,1);
        end
        %mu and V are calculated such that
        %f(x_j|x_i,theta.sigma2*T.t(i)) = phi(x_j,mu,V).
        mu = theta.mu_r + (X(i,:) - theta.mu_r)*(theta.sigma2_r + theta.sigma2 *p(j))/(theta.sigma2_r + theta.sigma2 *p(i));
        V = theta.sigma2 * T.t(i) * (theta.sigma2_r + theta.sigma2 *p(j))/(theta.sigma2_r + theta.sigma2 *p(i));
        weight = findWeightsSet(X(j,:),mu,V);
        for k = 0:N
            G(i,k+1) = sum(weight(k+1,:).*G(j,:).*F(v,:));
        end
    end
    for i=(n+1):(2*n-1)
        u = T.children(i,1);
        v = T.children(i,2);
        Vi = theta.sigma2_r + p(i)*theta.sigma2;
        %posterior density
        P(ch,i,:) = 1/like(ch) * F(u,:).*F(v,:).*G(i,:).*phi(X(i,:),theta.mu_r,Vi);
        %posterior probability that the liability at node i is
        %positive.
        P2(ch,i) = posterior((F(u,:).*F(v,:).*G(i,:))/like(ch),X(i,:),theta.mu_r,Vi);
    end
    Ps(1).array = P;
    Ps(2).array = P2;
    Ps(3).array = X;
    if like(ch)>0
        logL = logL + log(like(ch));
    else
        logL = -Inf;
    end
end
end

function w = findWeights(X,mu,sig2)
    %weights calculated for approximating \int phi(x,mu,sig2) f(x) dx
    N = length(X)-1;
    w = zeros(size(X)); %weight array
    h = X(2) - X(1); %spacing between two consecutive points.
    f = zeros(3,(N/2)+1); %matrix of antiderivatives used to calculate the weights.
    sig = sqrt(sig2);
    %f(1,i) = antiderivative of phi(x,mu,sig2) at x = X(2*i - 1)
    f(1,:) = .5*erf((X(1:2:(N+1))-mu)/(sqrt(2*sig2)));
    %f(2,i) = antiderivative of x*phi(x,mu,sig2) at x = X(2*i - 1)
    f(2,:) = -sig2*phi(X(1:2:(N+1)),mu,sig2) + mu*f(1,:);
    %f(3,i) = antiderivative of x^2*phi(x,mu,sig2) at x = X(2*i - 1)
    f(3,:) = (-2*sig*((X(1:2:(N+1))+mu).*exp(-((X(1:2:(N+1))-mu).^2)/(2*sig2))) + 2*sqrt(2*pi)*(mu^2 + sig2)*f(1,:))/(2*sqrt(2*pi));
    x = X(1:2:N);
    x2 = x.^2;
    f2 = f(:,2:end) - f(:,1:end-1);
    w(1:2:end) = (.5*(((2*h^2)+3*h*[x,0]+[x2,0]).*[f2(1,:),0] + ([0,x].*([0,x+h])).*[0,f2(1,:)]) - .5*(((3*h)+(2*[x,0])).*[f2(2,:),0] + ((2*[0,x])+h).*[0,f2(2,:)]) + .5*([f2(3,:),0] + [0,f2(3,:)]))/(h^2);
    w(2:2:end) = (-((x2+2*h*x).*f2(1,:))+(2*(x+h).*f2(2,:))-f2(3,:))/(h^2);
end


function w = findWeightsSet(X,mu,sig2)
    %a version of findWeights that find weight arrays for multiple values
    %of mu.
    N = length(X)-1;
    w = zeros(N+1); %weight matrix
    h = X(2) - X(1); %spacing between two consecutive points of X.
    sig = sqrt(sig2);
    x = repmat(X(1:2:(N+1)),N+1,1);
    MU = repmat(mu',1,(N/2)+1);
    %f1 = antiderivative of phi(x,mu,sig2) at x = X(2*i - 1)
    f1 = .5*erf((x-MU)/(sqrt(2*sig2)));
    %f2 = antiderivative of x*phi(x,mu,sig2) at x = X(2*i - 1)
    f2 = -sig2*phi(x,MU,sig2) + MU.*f1;
    %f3 = antiderivative of x^2*phi(x,mu,sig2) at x = X(2*i - 1)
    f3 = (-2*sig*((x+MU).*exp(-((x-MU).^2)/(2*sig2))) + 2*sqrt(2*pi)*(MU.^2 + sig2).*f1)/(2*sqrt(2*pi));
    x = x(:,1:(end-1));
    x2 = x.^2;
    F1 = f1(:,2:end) - f1(:,1:end-1);
    F2 = f2(:,2:end) - f2(:,1:end-1);
    F3 = f3(:,2:end) - f3(:,1:end-1);
    z = zeros(N+1,1);
    w(:,1:2:end) = (.5*(((2*h^2)+3*h*[x,z]+[x2,z]).*[F1,z] + ([z,x].*([z,x+h])).*[z,F1]) - .5*(((3*h)+(2*[x,z])).*[F2,z] + ((2*[z,x])+h).*[z,F2]) + .5*([F3,z] + [z,F3]))/(h^2);
    w(:,2:2:end) = (-((x2+2*h*x).*F1)+(2*(x+h).*F2)-F3)/(h^2);
end

function w = findWeights2(X,mu,sig2,ind)
    %version of findWeights which calculates weights for approximating
    %\int_0^Inf phi(x,mu,sig2) f(x) dx
    %ind = index such that ind = 2*m + 1 for integer m, X(ind) <= 0, and
    %X(ind+2) > 0.
    N = length(X)-1;
    w = zeros(size(X));
    h = X(2) - X(1);
    f = zeros(3,((N-ind+1)/2)+1);
	X2 = [0,X((ind+2):2:end)]; %0 used in place of X(ind) so that integrals
                               %in f2 are only over positive liabilities.
    sig = sqrt(sig2);
    f(1,:) = .5*erf((X2-mu)/(sqrt(2*sig2)));
    f(2,:) = -sig2*phi(X2,mu,sig2) + mu*f(1,:);
    f(3,:) = (-2*sig*((X2+mu).*exp(-((X2-mu).^2)/(2*sig2))) + 2*sqrt(2*pi)*(mu^2 + sig2)*f(1,:))/(2*sqrt(2*pi));
    x = X(ind:2:N);
    x2 = x.^2;
    f2 = f(:,2:end) - f(:,1:end-1);
    w(ind:2:end) = (.5*(((2*h^2)+3*h*[x,0]+[x2,0]).*[f2(1,:),0] + ([0,x].*([0,x+h])).*[0,f2(1,:)]) - .5*(((3*h)+(2*[x,0])).*[f2(2,:),0] + ((2*[0,x])+h).*[0,f2(2,:)]) + .5*([f2(3,:),0] + [0,f2(3,:)]))/(h^2);
    w((ind+1):2:end) = (-((x2+2*h*x).*f2(1,:))+(2*(x+h).*f2(2,:))-f2(3,:))/(h^2);
end

function Prob = posterior(Pos,X,mu,v)
	Prob = 0;
    if (X(end) > 0) && (X(1) < 0) %if X has both negative and positive liabilities
        x = X(1:2:end);
        [~,loc] = max(x(x <= 0));
        Ind = 2*loc - 1;
        w = findWeights2(X,mu,v,Ind);
		Prob = sum(w.*Pos);
    elseif (X(1) >= 0) %if X is entirely non-negative
        w = findWeights(X,mu,v);
		Prob = sum(w.*Pos);
    end
    %Prob remains as 0 if X is entirely non-positive.
end


function y = phi(x,mu,sigma2)
%normal pdf
y = 1/sqrt(2*pi*sigma2) * exp(-(x - mu).^2/(2*sigma2));
end

function y = Phi(x,mu,sigma2)
%normal cdf
y = 1/2 * (1 + erf((x-mu)/sqrt(2*sigma2)));
end

function y = InvPhi(x,mu,sigma2)
%inverse error function
y = sqrt(2*sigma2)*erfinv(2*x-1)+mu;
end






